; Sinclair QL 8049 IPC 2K ROM                                 (dgw 2019/03/27)
;
; source suitable for ASM48 v0.4.1
; 15 clocks per machine cycle (11MHz /3 /5 = 1.3636uS)
; 32 machine cycles per timer tick (11MHz /3 /5 /32 = 43.6364uS)
; 625 bytes unused (programmed as 00h, NOP)
;=============================================================================
; Memory map : 128 bytes RAM
; ~~~~~~~~~~~~~~~~~~~~~~~~~~
; 00-07h  register set 0
; 08-17h  stack (8 x2)
; 18-1Fh  register set 1 - used during interrupt
;
;               serial                     sound
;               ~~~~~~                     ~~~~~
;           r0  iostat pointer
;           r1  -qbytes (current)
;           r2  qin (current)              unused
;           r3  number bits received       unused
;           r4  40h/80h (SER1/2)           unused
;           r5  accumulated bits           unused
;           r6  -qend (current)            unused
;           r7  reg A preserved            reg A preserved
;
; 20h     IPC status bits (ipcstat) :
;
;           b0 keyboard being pressed
;           b1 beep status 0=off
;           b2 sh/ctrl/alt pressed
;           b3 repeat key
;           b4 SER1 open/closed  (open=1)
;           b5 SER2 open/closed  (open=1)
;           b6 WP MDV
;           b7 timeout occured for baud rate (=1)
;
; 21h     timeout counter (inactivity for baud rate)
;
; 22h     IO status bits (iostat) :
;
;           b0 current active serial port, SER1=0, SER2=1
;           b1 current queue full (=1)
;           b2 process sound on next interrupt (=1)
;           b3 awaiting start bit (=1), received (=0)
;           b4 last speaker output state (0=off)
;           b5 sound to be turned off (=1)
;           b6 SER1 queue full 0= <20 bytes, 1= 20-23 bytes
;           b7 SER2 queue full 0= <20 bytes, 1= 20-23 bytes
;       
; 23h     active serial queue size (positive value)
; 24h     baud rate inactivity timeout, see table at X02db 
; 25-26h  random number (IPC command E)
; 27h     random additive every main loop
; 28h     sh/ctrl/alt combination key value, see table at X05c0
; 29h     current keycode (default FFh)
; 2Ah     key debounce
; 2Bh     keyboard buffer pointer (default 2Ch)
; 2C-39h  keyboard buffer (7 x2 bytes : ov/s/c/a & keycode)
; 3A-4Ah  beep parameters :
;
;           3Ah  pitch1 0-255 
;           3Bh  pitch2 0-255  (must be >= pitch1)
;           3Ch  time_step lsb  0-32767  (x 72uS)
;           3Dh      "     msb
;           3Eh  duration lsb   0-32767  (x 72uS)
;           3Fh      "    msb
;           40h  pitch_step 0-15 / wrap 0-15
;           41h  fuzzy 0-15 / random 0-15
;
;         beep working parameters :
;
;           42h  current_pitch (pure)
;           43h  fuzzy_pitch   (current_pitch + any fuzzy element)
;           44h  random_pitch  (fuzzy_pitch + any random element)
;           45h  elapsed_time_step lsb
;           46h     "    "     "   msb
;           47h  elapsed_duration lsb
;           48h     "        "    msb
;           49h  wrap_counter
;           4Ah  timing_adjustment
; 
; 4Bh     pointer to currently active serial queue
; 4Ch     pending bytes SER1 (negative value)
; 4Dh     pending bytes SER2 (negative value)
; 4Eh     pointer, next out to ZX8302 for SER1 (default 52h)
; 4Fh     pointer, next in from SER1 (default 52h)
; 50h     pointer, next out to ZX8302 for SER2 (default 69h)
; 51h     pointer, next in from SER2 (default 69h)
; 52-68h  queue for SER1 (23 bytes)
; 69-7Fh  queue for SER2 (23 bytes)
;--------------------------------------------------------------------------------
; Ports : (quasi-bidirectional: 300ohm low, 50Kohm high, 272nS to 5Kohm when set)
; ~~~~~~~
; P1.0  (out)    KBD0
; P1.1  (out)    KBD1
; P1.2  (out)    KBD2
; P1.3  (out)    KBD3
; P1.4  (out)    KBD4
; P1.5  (out)    KBD5
; P1.6  (out)    KBD6
; P1.7  (out)    KBD7
;
; P2.0  (in)     RxD (SER2 DTE) ANDed TxD (SER1 DCE)
; P2.1  (out)    speaker (low turns speaker off by shunting external 1K resistor)
; P2.2  (out)    IPL2/0 (68008)
; P2.3  (out)    IPL1 (68008)
; P2.4  (out)    CTS (SER1 DCE)
; P2.5  (out)    DTR (SER2 DTE)
; P2.6  (in)     WP MDV (unused)
; P2.7  (in/out) COMDATA (ZX8302)
;
; /WR   (out)    COMCTL (ZX8302)
; /INT  (in)     tied to P2.0 (SER1/2 receive)
; T1    (in)     BAUDx4 (ZX8302)
;
; F0    user flag - set if interrupts are serviced
; F1    user flag - set if serial active
;================================================================================

        .equ    ipcstat, 20h    ; IPC status bits
        .equ    bdcount, 21h    ; baud counter
        .equ    iostat, 22h     ; IO status bots
        .equ    bdrate, 24h     ; Baud rate, default 03h (9600)
        .equ    key, 29h        ; current key, default FFh
        .equ    kbptr, 2Bh      ; keyboard buffer pointer, default 2Ch
        .equ    kbuff, 2Ch      ; keyboard buffer start
        .equ    q1bytes, 4Ch    ; pending bytes SER1
        .equ    q2bytes, 4Dh    ; pending bytes SER2
        .equ    q1out, 4Eh      ; out SER1 to ZX8302
        .equ    q1in, 4Fh       ; in from SER1
        .equ    q2out, 50h      ; out SER2 to ZX8302
        .equ    q2in, 51h       ; in from SER2

        .org    0

; reset vector

X0000:  jmp     X0009         
        nop

; external interrupt vector (on low level from SER1 TxD or SER2 RxD)

X0003:  jmp     X00de
        nop
        nop

; timer/counter interrupt vector

X0007:  jmp     X0100

; initialisation (also command 0 - reset IPC)

X0009:  stop    tcnt            ; stop timer/counter
        dis     i               ; disable external interrupts (serial)
        dis     tcnti           ; disable timer interrupts (beep)
        clr     f0
        cpl     f0              ; set user flag 0 (do service interrupts) 
        orl     p2,#1           ; P2.0 SER1/2 receive is input
        anl     p2,#0FDh        ; P2.1 speaker low (off)
        orl     p2,#04          ; P2.2 IPL2/0 high
        orl     p2,#08          ; P2.3 IPL1 high
        orl     p2,#10h         ; P2.4 SER1 CTS high
        orl     p2,#20h         ; P2.5 SER2 DTR high
        orl     p2,#40h         ; P2.6 WP MDV high (unused)
        orl     p2,#80h         ; P2.7 COMDATA high
        mov     r0,#20h
        mov     r2,#60h
X0022:  mov     @r0,#0          ; workspace 20-7Fh cleared
        inc     r0
        djnz    r2,X0022
        mov     r0,#key
        mov     @r0,#0FFh       ; default keycode (none)
        mov     r0,#kbptr
        mov     @r0,#2Ch        ; empty keyboard buffer position
        mov     r0,#bdrate
        mov     @r0,#03         ; default baud 9600
        mov     r0,#q1out
        mov     @r0,#52h        ; SER1 queue empty
        inc     r0
        mov     @r0,#52h
        inc     r0
        mov     @r0,#69h        ; SER2 queue empty
        inc     r0
        mov     @r0,#69h
        clr     f1              ; clear user flag 1 (no serial activity)
        call    X004a           ; initial housekeeping task call
X0043:  in      a,p2
        anl     a,#80h          ; COMDATA high ?
        jz      X0043           ; no, wait until high
        jmp     X0500           ; perform main loop: scan keyboard, serial
                                ; receive or sound, IPC command, and repeat

; housekeeping task - sound/serial

X004a:  stop    tcnt            ; stop timer/counter
        sel     rb1             ; switch to register bank 1
        mov     r0,#ipcstat
        mov     a,@r0
        anl     a,#2            ; beep status ?
        mov     r0,#iostat
        jz      X005d           ; was off, so do serial
        mov     a,@r0
        orl     a,#4            ; flag 'process sound next interrupt'
        anl     a,#0efh         ; speaker output status to off
        mov     @r0,a           ; update iostat
        jmp     X03e4           ; toggle speaker, set timer for pitch and return

; check if serial queues nearing capacity

X005d:  mov     r2,#0           ; assume both SER1/2 not full
        mov     r0,#q1bytes     ; -pending bytes SER1
        mov     a,@r0
        cpl     a
        inc     a               ; make value positive
        anl     a,#3fh          ; restrict 0-31 (buffer only 23)
        add     a,#0ech         ; -20, almost full ?
        jnc     X006e           ; no, < 20
        mov     a,r2
        orl     a,#10h          ; else flag SER1 full (almost)
        mov     r2,a
X006e:  inc     r0
        mov     a,@r0           ; -pending bytes SER2
        cpl     a
        inc     a               ; make value positive
        anl     a,#3fh          ; restrict 0-31 (buffer only 23)
        add     a,#0ech         ; -20, almost full ?
        jnc     X007c           ; no, < 20 
        mov     a,r2
        orl     a,#20h          ; else flag SER2 full (almost)
        mov     r2,a
X007c:  mov     a,r2
        cpl     a               ; invert meaning, bit set if not full 
        mov     r0,#ipcstat
        anl     a,@r0           ; corresponding bit set if SER1/2 also open
        mov     r0,#iostat
        jb4     X009f           ; jump if SER1 open & not full
        jb5     X0091           ; jump if SER2 open & not full
        stop    tcnt            ; stop timer/counter
        mov     a,#0            ; 256 ticks @ 11MHz/3/5/32
        mov     t,a             ; reload timer
        dis     tcnti           ; disable timer interrupts
        en      tcnti           ; enable timer interrupts
        strt    t               ; start timer (next interrupt in 11.1708mS)
        mov     a,r7            ; (restore A, if call was from interrupt)
        sel     rb0             ; restore to register bank 0
        retr                    ; restore status and return

; SER2 open & not full, prepare to receive next byte

X0091:  mov     a,@r0
        orl     a,#1            ; make SER2 current port
        mov     @r0,a
        mov     r2,#q2in        ; queue in SER2
        mov     r1,#q2bytes     ; -pending bytes SER2
        mov     r6,#80h         ; -end of queue SER2 (80h)
        mov     r4,#80h         ; port bitmask
        jmp     X00b4

; SER1 open & not full, prepare to receive next byte

X009f:  jb5     X00af           ; jump if SER2 also open & not full
X00a1:  mov     a,@r0
        anl     a,#0feh         ; make SER1 current port
        mov     @r0,a
        mov     r2,#q1in        ; queue in SER1
        mov     r1,#q1bytes     ; -pending bytes SER1
        mov     r6,#97h         ; -end of queue SER1 (69h)
        mov     r4,#40h         ; port bitmask
        jmp     X00b4

; both SER1 & SER2 open & not full, so switch between on each call

X00af:  mov     a,@r0           ; which is current SER port ?
        jb0     X00a1           ; SER2, so now use SER1
        jmp     X0091           ; else SER1 so now use SER2

; set up registers (bank 1) for current serial port

X00b4:  mov     a,r4            ; port bitmask
        cpl     a               ; (0BFh SER1, 7Fh SER2)
        anl     a,@r0           ; clear queue full bit for SER port
        anl     a,#0fdh         ; clear flag 'current queue full'
        orl     a,#08           ; set flag 'awaiting start bit'
        mov     @r0,a           ; update iostat
        dec     r0              ; point to baud limit counter
        mov     @r0,#0          ; reset it
        inc     r0
        inc     r0              ; point to active queue size
        mov     @r0,#0          ; reset it
        mov     r0,#4bh         ; point to current serial queue pointer
        mov     a,r2            ; which will be this queue
        mov     @r0,a           ; update
        mov     r0,a
        mov     a,@r0           ; get current queue in pointer
        mov     r2,a            ; preserve
        mov     r0,#iostat
        cpl     f1              ; set user flag 1 (serial activity)
        mov     r3,#0           ; reset number of bits received
        en      i               ; enable external interrupts (for start bit)
        dis     tcnti           ; disable timer/counter interrupts
        mov     a,r4            ; current port ?
        jb7     X00d9           ; is SER2
        mov     a,r7            ; (restore A, if call was from interrupt)
        sel     rb0             ; restore to register bank 0
        anl     p2,#0efh        ; SER1 CTS output low
        retr                    ; restore status and return

X00d9:  mov     a,r7            ; (restore A, if call was from interrupt)
        sel     rb0             ; restore to register bank 0
        anl     p2,#0dfh        ; SER2 DTE output low
        retr                    ; restore status and return

; external interrupt handler (on start bit from SER1 or SER2)

X00de:  jf0     X00e1           ; if user flag 0 set, service interrupt
        retr                    ; else just restore status and return

X00e1:  sel     rb1             ; switch to register bank 1
        mov     r7,a            ; (preserve A)
        mov     a,#0ffh         ; -1 (count just one interrupt)
        mov     t,a             ; load counter
        en      tcnti           ; enable counter interrupts
        strt    cnt             ; start counter (T1 input = BAUDx4)
        dis     i               ; disable external interrupts
        mov     a,r7            ; (restore A)
        retr                    ; restore status and return

        .org    100h            ; EB-FFh = 21 bytes spare

; timer/counter interrupt handler

X0100:  jf0     X0103           ; if user flag 0 set, service interrupts
        retr                    ; else just restore status and return

X0103:  sel     rb1             ; switch to register bank 1
        mov     r7,a            ; (preserve A)
        stop    tcnt            ; stop timer/counter
        mov     a,#0fch         ; -4 (count BAUDx4) for next bit
        mov     t,a             ; load timer (assume serial activity)
        strt    cnt             ; start counter (T1 input = BAUDx4)
        jf1     X0137           ; jump if user flag 1 set (serial active)
        mov     r0,#iostat
        mov     a,@r0
        anl     a,#4            ; process sound ?
        jnz     X0115           ; yes
        jmp     X004a           ; else do housekeeping task
                                ; and return from interrupt state
X0115:  jmp     X0327

; command 2 & 3  - open SER1 or SER2

X0117:  mov     r0,#ipcstat
        jb0     X0121           ; jump if SER2
        mov     a,@r0
        orl     a,#10h          ; signal SER1
        mov     @r0,a           ; has been opened
        jmp     X0500           ; back to main loop

X0121:  mov     a,@r0
        orl     a,#20h          ; signal SER2
        mov     @r0,a           ; has been opened
        jmp     X0500           ; back to main loop

; command 4 & 5  - close SER1 or SER2

X0127:  mov     r0,#ipcstat
        jb0     X0131           ; jump if SER2
        mov     a,@r0
        anl     a,#0efh         ; signal SER1
        mov     @r0,a           ; has been closed
        jmp     X0500           ; back to main loop

X0131:  mov     a,@r0
        anl     a,#0dfh         ; signal SER2
        mov     @r0,a           ; has been closed
        jmp     X0500           ; back to main loop

; interrupt routine for serial, assemble byte

X0137:  in      a,p2
        anl     a,#1            ; serial receive data
        xch     a,r3            ; any bits received yet ?
        jnz     X0155           ; yes
        xch     a,r3            ; else first, is it low start bit ?
        jnz     X014c           ; no, false start bit
        mov     a,@r0           ; (iostat)
        anl     a,#0f7h         ; clear flag 'awaiting start bit'
        mov     @r0,a
        dec     r0
        mov     @r0,#0          ; reset baud limit counter
        inc     r0              ; (restore pointer to iostat)
        mov     r3,#9           ; bits to receive (including stop bit)
        mov     a,r7            ; (restore A)
        retr                    ; restore status and return

X014c:  stop    tcnt            ; stop timer/counter
        dis     tcnti           ; disable timer interrupts
        en      i               ; enable external interrupts
        mov     a,@r0
        orl     a,#8            ; set flag 'awaiting start bit'
        mov     @r0,a
        mov     a,r7            ; (restore A)
        retr                    ; restore status and return

X0155:  xch     a,r3            ; (restore bit counter)
        djnz    r3,X018a        ; jump until all bits received

        stop    tcnt            ; stop timer/counter
        dis     tcnti           ; disable timer/counter interrupts
        en      i               ; enable external interrupts
        jnz     X0164           ; jump if low stop bit received OK
        mov     a,@r0           ; (iostat)
        anl     a,r4            ; current port queue full ?
        jnz     X0184           ; yes (ignore bad stop bit)
        mov     a,r4
        orl     a,@r0           ; else in event of bad stop bit
        mov     @r0,a           ; set current queue as full

X0164:  mov     a,r2            ; current queue pointer
        mov     r0,a
        mov     a,r5            ; accumulated bits
        mov     @r0,a           ; put newly received byte into queue
        mov     a,@r1           ; get queue bytes counter
        dec     a               ; (negative count value)
        mov     @r1,a           ; update
        inc     r0              ; point to next queue location
        mov     a,r0
        add     a,r6            ; carry set if reached queue end
        mov     a,r0            ; (current position)
        jnc     X0173           ; queue not at end
        add     a,#0e9h         ; else -23 back to queue start
X0173:  mov     r2,a            ; update current queue pointer
        mov     r0,#iostat      ; (restore)
        mov     a,@r1           ; queue bytes counter
        add     a,#14h          ; queue completely full (23 bytes) ?
        jnz     X0181           ; no
        orl     p2,#30h         ; else busy both SER1 CTS & SER2 DTR
        mov     a,@r0
        orl     a,#2            ; flag current queue full
        mov     @r0,a
X0181:  inc     r0              ; point to active queue size
        inc     @r0             ; increment it
        dec     r0              ; (point back at iostat)
X0184:  mov     a,@r0
        orl     a,#08           ; set flag 'awaiting start bit'
        mov     @r0,a
        mov     a,r7            ; (restore A)
        retr                    ; restore status and return

X018a:  rrc     a               ; bit received to carry
        mov     a,r5            ; bits accumulated so far
        rrc     a
        mov     r5,a            ; add new bit
        mov     a,r7            ; (restore A)
        retr                    ; restore status and return

        .org    200h            ; 190-1FFh = 112 bytes spare

; get IPC command

X0200:  in      a,p2
        anl     a,#80h          ; COMDATA low ?
        jz      X0207           ; yes, get command
        jmp     X0500           ; else back to main loop

X0207:  call    X074f           ; fetch 4 bit command 0-F
        add     a,#0ch          ; add address offset  (020Ch)
        jmpp    @a              ; into table below

; command jump table (holds lower 8 bits for PC, upper = 02h)

X020c:  .db     #1Ch            ; 0 reset IPC
        .db     #28h            ; 1 report input status
        .db     #1Eh            ; 2 open SER1
        .db     #1Eh            ; 3 open SER2
        .db     #20h            ; 4 close SER1
        .db     #20h            ; 5 close SER2
        .db     #89h            ; 6 read SER1
        .db     #89h            ; 7 read SER2
        .db     #4Dh            ; 8 read keyboard
        .db     #76h            ; 9 keyrow
        .db     #22h            ; A set sound (beep)
        .db     #24h            ; B kill sound (beep)
        .db     #0C3h           ; C set P2.3
        .db     #0CFh           ; D set baud rate
        .db     #0E3h           ; E random number
        .db     #0F0h           ; F test

X021c:  jmp     X0009           ; reset IPC
X021e:  jmp     X0117           ; open SER 1/2
X0220:  jmp     X0127           ; close SER 1/2
X0222:  jmp     X0300           ; set sound (beep)
X0224:  jmp     X031f           ; kill sound (beep)
X0226:  jmp     X0500           ; main loop (never called)

; command 1 - report input status

X0228:  in      a,p2
        anl     a,#40h          ; get latest value of P2.6 (WP MDV)
        mov     r0,#ipcstat
        orl     a,@r0           ; update status of P2.6 (WP MDV)
        anl     a,#0cfh         ; clear SER status bits 4 & 5 (no pending input)
        jb3     X0234           ; jump if repeat key active
        jmp     X0236

X0234:  orl     a,#1            ; signal key being pressed (set bit 0)
X0236:  mov     r2,a            ; preserve result
        mov     r0,#q1bytes
        mov     a,@r0           ; SER1 pending bytes ?
        jz      X0240           ; none
        mov     a,r2
        orl     a,#10h          ; bit 4, set SER1 byte pending
        mov     r2,a
X0240:  inc     r0
        mov     a,@r0           ; SER2 pending bytes ?
        jz      X0248           ; none
        mov     a,r2
        orl     a,#20h          ; bit 5, set SER2 byte pending
        mov     r2,a
X0248:  mov     a,r2            ; result to be returned
        call    X0784           ; send 8 bit data
        jmp     X0500           ; back to main loop

; command 8 - read keyboard

X024d:  mov     r0,#kbptr       ; current buffer pointer
        mov     a,@r0
        add     a,#0d4h         ; form length of entries
        clr     c               ; clear carry
        rrc     a               ; form actual number
        mov     r2,a            ; of keystrokes in buffer
        mov     r1,#ipcstat
        mov     a,@r1
        anl     a,#08           ; isolate repeat key status
        orl     a,r2            ; add number of keys in buffer
        call    X0789           ; send 4 bit data
        mov     a,r2
        jz      X026e           ; buffer empty
        mov     r1,#kbuff       ; start of buffer
X0262:  mov     a,@r1           ; get key ?/sh/ctrl/alt
        call    X0789           ; send 4 bit data
        inc     r1
        mov     a,@r1           ; get keycode
        call    X0784           ; send 8 bit data
        inc     r1
        djnz    r2,X0262        ; until buffer read
        mov     @r0,#kbuff      ; reset buffer pointer
X026e:  mov     r0,#ipcstat
        mov     a,@r0
        anl     a,#0feh         ; clear keyboard being pressed bit
        mov     @r0,a
        jmp     X0500           ; back to main loop

; command 9 - keyrow

X0276:  call    X074f           ; fetch 4 bit data
        anl     a,#7            ; restrict to 0-7
        inc     a               ; +1
        mov     r2,a            ; as loop counter
        mov     a,#0            ; prepare scan mask
        clr     c               ; clear carry
        cpl     c               ; set carry
X0280:  rlc     a               ; rotate 1 into
        djnz    r2,X0280        ; keyrow bit position
        call    X05c8           ; scan keyboard matrix row
        call    X0784           ; send 8 bit data
        jmp     X0500           ; back to main loop

; command 6 & 7 - read SER1 or SER2

X0289:  jb0     X0293           ; jump if SER2
        mov     r0,#q1bytes     ; -pending bytes SER1
        mov     r4,#97h         ; -end of queue SER1
        mov     r5,#q1out       ; queue out SER1
        jmp     X0299

X0293:  mov     r0,#q2bytes     ; -pending bytes SER2
        mov     r4,#80h         ; -end of queue SER2
        mov     r5,#q2out       ; queue out SER2
X0299:  mov     a,@r0           ; -pending bytes
        cpl     a
        inc     a               ; form positive value
        anl     a,#3fh          ; restrict 0-31 (buffer only 23)
        mov     r3,a            ; preserve
        mov     r1,#iostat
        mov     a,@r1
        anl     a,#0c0h         ; isolate SER full bits
        orl     a,r3            ; add to length value
        call    X0784           ; send 8 bit data
        mov     a,r3            ; bytes to send
        jz      X02c1           ; none, finished
        mov     a,r5            ; queue pointer
        mov     r1,a
        mov     a,@r1           ; fetch byte
        mov     r1,a
X02ae:  mov     a,@r1
        call    X0784           ; send 8 bit data
        inc     @r0             ; decrement pending count
        inc     r1              ; progress queue pointer
        mov     a,r1
        add     a,r4            ; end of queue reached ?
        jnc     X02bb           ; no
        mov     a,r1            ; else wrap back around to start
        add     a,#0e9h         ; -23 queue length
        mov     r1,a
X02bb:  djnz    r3,X02ae        ; until all bytes sent
        mov     a,r5            ; queue out pointer
        mov     r0,a
        mov     a,r1            ; empty queue position
        mov     @r0,a           ; update queue out pointer
X02c1:  jmp     X0500           ; back to main loop

; command C - set P2.3

X02c3:  call    X074f           ; fetch 4 bit data
        jb0     X02cb
        anl     p2,#0f7h        ; clear P2.3 (IPL1)
        jmp     X0500           ; back to main loop
;
X02cb:  orl     p2,#8           ; set P2.3 (IPL1)
        jmp     X0500           ; back to main loop

; command D - set baud rate

X02cf:  call    X074f           ; fetch 4 bit data
        anl     a,#7            ; restrict to 0-7
        add     a,#0dbh         ; index into table
        movp    a,@a            ; get table entry
        mov     r0,#bdrate
        mov     @r0,a           ; update current setting
        jmp     X0500           ; back to main loop

; timeout values for inactivity before switching serial ports (if both open)

X02db:  .db     #01h            ; 19200
        .db     #03h            ;  9600
        .db     #07h            ;  4800
        .db     #0Fh            ;  2400
        .db     #1Fh            ;  1200
        .db     #3Fh            ;   600
        .db     #7Fh            ;   300
        .db     #0FFh           ;    75

; command E - random number

X02e3:  call    X073a           ; randomise IPC random number
        mov     r1,#25h         ; get it
        mov     a,@r1           ; random number lsb
        call    X0784           ; send 8 bit data
        inc     r1
        mov     a,@r1           ; random number msb
        call    X0784           ; send 8 bit data
        jmp     X0500           ; back to main loop

; command F - test

X02f0:  call    X077c           ; fetch 8 bit data
        call    X0784           ; send 8 bit data
        jmp     X0500           ; back to main loop

        .org    300h            ; 2F6-2FFh = 10 bytes spare

; command A - set sound (beep)

X0300:  mov     r0,#3ah         ; buffer for beep parameters
        mov     r1,#8           ; number of parameters
X0304:  call    X077c           ; fetch 8 bit data (or 4+4 data)
        mov     @r0,a           ; store them
        inc     r0
        djnz    r1,X0304        ; until done
        call    X0415           ; copy starting pitch to current_pitch
        inc     r0
        mov     @r0,a           ; also to fuzzy_pitch
        inc     r0
        mov     @r0,a           ; and random_pitch
        mov     r1,#5
X0312:  inc     r0
        mov     @r0,#0          ; clear other working values
        djnz    r1,X0312
        mov     r0,#ipcstat
        mov     a,@r0
        orl     a,#2            ; beep status to on
        mov     @r0,a
        jmp     X0500           ; back to main loop

; command B - kill sound (beep)

X031f:  mov     r0,#ipcstat
        mov     a,@r0
        anl     a,#0fdh         ; beep status to off
        mov     @r0,a
        jmp     X0500           ; back to main loop

; interrupt routine for sound

X0327:  stop    tcnt            ; stop timer/counter
        mov     r0,#ipcstat
        mov     a,@r0
        anl     a,#2            ; beep status ?
        jnz     X033b           ; is on
        mov     r0,#iostat
        mov     a,@r0
        anl     a,#0fbh         ; clear flag 'process sound on next interrupt'
        anl     a,#0dfh         ; clear flag 'sound to be turned off'
        mov     @r0,a
        anl     p2,#0fdh        ; speaker output off
        jmp     X004a           ; finish interrupt by doing housekeeping

X033b:  mov     r0,#3eh
        mov     a,@r0           ; is duration lsb zero ?
        jnz     X0344           ; no
        inc     r0
        mov     a,@r0           ; is duration msb also zero ?
        jz      X0365           ; yes, then beep is forever

; check for duration expiry

X0344:  mov     r0,#4ah         ; timing_adjustment
        inc     @r0
        inc     @r0             ; +2
        mov     r1,#47h         ; elapsed_duration lsb
        mov     r0,#44h         ; random_pitch
        mov     a,@r0
        jnz     X0351           ; 1 to 255 are true values
        mov     a,#0ffh         ; but 0 = 256, so just use 255
X0351:  call    X0716           ; add A to word at @r1
        mov     r0,#4ah         ; timing_adjustment
        mov     a,@r0
        call    X0716           ; add A to word at @r1
        mov     r0,#3fh         ; duration msb
        inc     r1              ; (elapsed_duration msb)
        call    X070b           ; has duration expired ?
        jnc     X0365           ; no
        mov     r0,#iostat
        mov     a,@r0
        orl     a,#20h          ; else set flag 'sound to be turned off'
        mov     @r0,a

; check for time_step expiry

X0365:  mov     r1,#45h         ; elapsed_time_step lsb
        mov     r0,#44h         ; random_pitch
        mov     a,@r0
        jnz     X036e           ; 1 to 255 are true values
        mov     a,#0ffh         ; but 0 = 256, so just use 255
X036e:  call    X0716           ; add A to word at @r1
        mov     r0,#4ah         ; timing_adjustment
        mov     a,@r0
        call    X0716           ; add A to word at @r1
        mov     r0,#3dh         ; time_step msb
        inc     r1              ; (elapsed_time_step msb)
        call    X070b           ; has time_step expired ?
        jnc     X03e4           ; no, done (skip pitch_step, wrap and fuzzy)
        mov     r0,#4ah         ; timing_adjustment
        mov     a,@r0
        add     a,#3            ; +3
        mov     @r0,a
        mov     r0,#45h         ; reset elapsed_time_step counter
        mov     @r0,#0
        inc     r0
        mov     @r0,#0

; time_step has expired so perform pitch_step change

        mov     r0,#42h         ; current_pitch
        mov     r1,#40h         ; pitch_step / wrap
        mov     a,@r1
        swap    a
        orl     a,#0f0h         ; isolate pitch_step
        jb3     X0395           ; jump if 8-15 (byte negative)
        anl     a,#0fh          ; else make byte positive
X0395:  add     a,@r0           ; + current_pitch
        mov     @r0,a           ; update current_pitch

; now check if current_pitch within bounds of pitch1 and pitch2

        mov     r1,#3ah
        call    X0700           ; is current_pitch <= pitch1 ?
        jc      X03a5           ; yes, out of bounds or at boundary
        mov     r1,#3bh
        call    X0700           ; is current_pitch = pitch2 ?
        jz      X03a5           ; yes, at boundary
        jc      X03d3           ; else jump if OK (current_pitch < pitch2)

; perform wrap (repeat sweep)

X03a5:  mov     r0,#4ah         ; timing_adjustment
        inc     @r0
        inc     @r0             ; +2
        mov     r0,#40h         ; pitch_step / wrap
        mov     a,@r0
        mov     r0,#49h         ; (wrap_counter)
        orl     a,#0f0h         ; isolate wrap parameter
        inc     a               ; was wrap 15 (forever) ?
        jz      X03bd           ; yes
        dec     a               ; else redress wrap value
        anl     a,#0fh          ; make positive
        jz      X03c2           ; jump if wrap was 0 (never)
        cpl     a
        inc     a               ; two's complement
        add     a,@r0           ; + wrap_counter
        jc      X03c2           ; jump if wrap expired
X03bd:  inc     @r0             ; update wrap_counter
        call    X0415           ; and reset current_pitch to starting pitch
        jmp     X03d3           ; do fuzzy

; wrap expired, reverse pitch_step direction

X03c2:  mov     @r0,#0          ; reset wrap_counter
        mov     r1,#40h         ; pitch_step / wrap parameters
        mov     a,@r1
        swap    a
        cpl     a
        inc     a               ; two's complement
        anl     a,#0fh          ; -pitch_step (reverse step direction)
        swap    a
        mov     r2,a
        mov     a,@r1
        anl     a,#0fh          ; isolate wrap parameter
        orl     a,r2            ; -pitch_step / wrap
        mov     @r1,a           ; update

; do fuzzy

X03d3:  mov     r1,#41h         ; fuzzy / random parameters
        mov     a,@r1
        swap    a               ; pretend fuzzy parameter is random
        mov     @r1,a           ; (temporarily)
        call    X0422           ; generate random sound element
        mov     r1,#42h         ; current_pitch
        add     a,@r1
        inc     r1
        mov     @r1,a           ; update fuzzy_pitch
        mov     r1,#41h         ; fuzzy / random parameters (switched)
        mov     a,@r1
        swap    a               ; restore fuzzy / random parameters
        mov     @r1,a

; toggle speaker output

X03e4:  mov     r0,#iostat
        mov     a,@r0
        anl     a,#10h          ; last speaker state ?
        jz      X03f3           ; was off
        anl     p2,#0fdh        ; else speaker output to off
        mov     a,@r0
        anl     a,#0efh
        mov     @r0,a           ; update speaker status to off
        jmp     X0400

X03f3:  orl     p2,#2           ; speaker output to on
        mov     a,@r0
        orl     a,#10h          ; update speaker status to on
        mov     @r0,a
        jmp     X0400

        .org    400h            ; 3FB-3FFh = 5 bytes spare

; prepare timer for sound (add +343uS to timer value for true frequency*2)

X0400:  call    X0422           ; generate random sound element
        mov     r1,#43h         ; fuzzy_pitch
        add     a,@r1           ; add random element (or zero)
        inc     r1
        mov     @r1,a           ; update random_pitch
        cpl     a
        inc     a               ; two's complement
        stop    tcnt            ; stop timer/counter
        mov     t,a             ; load timer (each tick 43.6364uS)
        dis     tcnti           ; disable timer interrupts
        en      tcnti           ; enable timer interrupts
        strt    t               ; start timer
        mov     r0,#4ah         ; timing_adjustment
        mov     @r0,#4          ; initial value
        mov     a,r7            ; (restore A, if called from interrupt)
        sel     rb0             ; restore to register bank 0
        retr                    ; restore status and return

; set current_pitch as pitch1 or pitch2 based on pitch_step direction

X0415:  mov     r0,#40h         ; pitch_step / wrap
        mov     a,@r0
        mov     r1,#3bh         ; use pitch2 parameter
        jb7     X041d           ; if pitch_step is 8-15 (-7 to -1)
        dec     r1              ; else use pitch1 (0 to 7)
X041d:  mov     r0,#42h
        mov     a,@r1           ; set current_pitch
        mov     @r0,a
        ret

; generate random sound element

X0422:  mov     r0,#41h         ; fuzzy / random
        mov     a,@r0
        jb3     X0429           ; jump if random 8-15
        clr     a               ; else 0-7 ignored, no randomness
        ret

X0429:  mov     r0,#4ah         ; timing_adjustment
        mov     a,@r0
        add     a,#7            ; +7
        mov     @r0,a
        call    X073a           ; randomise IPC random number
        mov     r0,#41h         ; beep fuzzy / random parameters
        mov     r1,#25h         ; current IPC random number
        mov     a,@r0
        anl     a,#7            ; isolate beep random parameter
        inc     a
        mov     r2,a            ; as counter
        clr     a
        clr     c               ; start loop with carry set
X043c:  cpl     c
        rlc     a               ; create bitmask
        djnz    r2,X043c
        anl     a,@r1           ; form final random value
        ret

        .org    500h            ; 441-4FF = 191 bytes spare

; scan keyboard

X0500:  mov     r0,#27h
        inc     @r0             ; increment random additive
        mov     r2,#8           ; (starting keyrow number+1)
        mov     a,#80h          ; mask for s/c/a keyrow
        call    X05c8           ; scan keyboard matrix row
        mov     r6,a            ; preserve column data
        anl     a,#7            ; isolate s/c/a keys
        add     a,#0c0h         ; offset into table at X05c0
        movp    a,@a            ; get combination key value
        mov     r4,a            ; preserve value
        mov     r0,#28h
        xrl     a,@r0           ; same as previous ?
        jz      X0521           ; yes, look for normal key
        mov     r1,#ipcstat
        mov     a,@r1
        orl     a,#4            ; set s/c/a key flag
        anl     a,#0f7h         ; clear repeat key flag
        mov     @r1,a           ; update ipcstat flags
        mov     a,r4            ; retrieve value
        mov     @r0,a           ; update s/c/a key setting
        jmp     X05be           ; end this call by doing any sound/serial

X0521:  mov     a,r6            ; retrieve column data for keyrow 7
        anl     a,#0f8h         ; exclude s/c/a keys
        mov     r6,#0ffh        ; default keycode for no key
        call    X05d3           ; convert to keycode
        jc      X053f           ; jump if multiple keys
        dec     r2              ; decrement current keyrow number
X052b:  mov     a,r3            ; (previous keyrow mask)
        rr      a               ; advance to next keyrow
        call    X05c8           ; scan keyboard matrix row
        call    X05d3           ; convert to keycode
        jc      X053f           ; jump if multiple keys
        djnz    r2,X052b        ; until all keyrows scanned
        mov     a,r6
        inc     a               ; valid keycode ?
        jnz     X054e           ; yes
        mov     r0,#ipcstat
        mov     a,@r0
        anl     a,#0fbh         ; clear s/c/a key flag
        mov     @r0,a
X053f:  mov     r0,#ipcstat
        mov     a,@r0
        anl     a,#0f7h         ; clear repeat key flag
        mov     @r0,a
        mov     r0,#key
        mov     @r0,#0ffh       ; no key currently being pressed
        inc     r0
        mov     @r0,#0          ; reset debounce
        jmp     X05be           ; now go do any sound/serial

X054e:  mov     r0,#ipcstat
        mov     a,@r0
        anl     a,#4            ; s/c/a status ?
        jnz     X05be           ; if on, go do any sound/serial
        mov     r0,#key
        mov     a,@r0
        xrl     a,r6            ; same keycode as last ?
        jz      X0562           ; yes
        mov     a,r6
        mov     @r0,a
        inc     r0
        mov     @r0,#0          ; reset key debounce
        jmp     X05be           ; now go do any sound/serial

X0562:  inc     r0
        mov     a,@r0           ; get key debounce
        add     a,#0fch         ; =4 ?
        jz      X056d           ; yes
        jc      X05b8           ; if >4 flag repeat key
        inc     @r0             ; else increment key debounce
        jmp     X05be           ; now go do any sound/serial

X056d:  inc     @r0             ; increment key debounce
        mov     r0,#28h
        mov     a,@r0           ; last keycode
        add     a,#0fdh         ; =3 'ctrl + alt' ?
        jnz     X058d           ; no 
        mov     a,r6            ; current keycode
        add     a,#0c1h         ; =63 '7' ?
        jnz     X058d           ; no 

; 'ctrl + alt + 7' pressed

        anl     p2,#0fbh        ; assert 68008 IPL2/0 pin
        dis     i               ; disable external interrupts
        dis     tcnti           ; disable timer interrupts
        stop    tcnt            ; stop timer/counter
        orl     p2,#10h         ; busy SER1 CTS
        orl     p2,#20h         ; busy SER2 DTR
        mov     r0,#0ffh
X0585:  mov     r1,#0ffh
X0587:  djnz    r1,X0587
        djnz    r0,X0585        ; wait 0.3547 second
        jmp     X0009           ; reset IPC

; put keycode into buffer

X058d:  mov     r0,#kbptr       ; keyboard buffer pointer
        mov     a,@r0
        mov     r1,a
        add     a,#0c6h         ; =$3A buffer full ?
        jz      X05b0           ; yes
        mov     r0,#28h
        mov     a,@r0           ; s/c/a combination keycode
        mov     @r1,a           ; into buffer
        inc     r1
        mov     a,r6            ; current keycode
        mov     @r1,a           ; into buffer
        inc     r1
        mov     a,r1
        mov     r0,#kbptr
        mov     @r0,a           ; update keyboard buffer pointer
        mov     r0,#ipcstat
        mov     a,@r0
        orl     a,#1            ; set key being pressed flag
        mov     @r0,a
        mov     r0,#27h         ; random additive
        mov     a,@r0
        mov     r1,#25h         ; random number location
        call    X0716           ; add A to word at @r1
        jmp     X05be           ; now go do any sound/serial

; keyboard buffer full

X05b0:  mov     r0,#38h         ; last keyboard buffer entry position
        mov     a,@r0
        orl     a,#8            ; signal overflow occured
        mov     @r0,a
        jmp     X05be           ; now go do any sound/serial

; repeat key

X05b8:  mov     r0,#ipcstat
        mov     a,@r0
        orl     a,#8            ; set repeat key flag
        mov     @r0,a
X05be:  jmp     X0600           ; deal with any sound/serial activity

; sh/ctrl/alt key table

X05c0:  .db     #00h            ; none
        .db     #04h            ; sh
        .db     #02h            ; ctrl
        .db     #06h            ; sh + ctrl
        .db     #01h            ; alt
        .db     #05h            ; sh + alt
        .db     #03h            ; ctrl + alt
        .db     #07h            ; sh + ctrl + alt

; scan keyboard matrix row (single keyrow bit set in A)

X05c8:  outl    p1,a            ; output row mask
        mov     r3,a            ; preserve (used in subsquent calls)
        movx    a,@r0           ; ensure Bus/P0 is in floating state
        nop                     ; (see MCS-48 page 60 section 6.6) pulse RD low
        nop
        nop
        nop                     ; propagation delay
        nop
        nop
        in      a,p0            ; read column data for row (ins a,bus)
        retr                    ; restore status and return

; convert keyrow scan to valid keycode for single key pressed (else carry set)

X05d3:  jz      X05f9           ; no key pressed in this keyrow
        xch     a,r6            ; current keycode
        inc     a               ; no key pressed so far ?
        jz      X05dc           ; yes, process new key
X05d9:  clr     c               ; else clear carry
        cpl     c               ; set carry, signals multiple keys
        ret

X05dc:  mov     r4,#8           ; bits to scan
        mov     a,r6            ; keyrow column data
X05df:  jb7     X05e8           ; jump if key in column pressed
        rl      a
        djnz    r4,X05df        ; until all columns done
        mov     r6,#0ffh        ; keycode for no key
        clr     c               ; clear carry
        ret

X05e8:  anl     a,#7fh          ; more than one key on column ?
        jnz     X05d9           ; yes, reject as invalid
        mov     a,r2            ; keyrow number+1
        dec     a               ; true keyrow number
        cpl     a
        anl     a,#7
        rl      a
        rl      a
        rl      a               ; form row position
        mov     r6,a
        mov     a,r4
        dec     a               ; form column position
        orl     a,r6
        mov     r6,a            ; keycode result
X05f9:  clr     c               ; clear carry (valid keycode)
        ret

        .org    600h            ; 5FB-5FFh = 5 bytes spare

; deal with any sound/serial activity

X0600:  jf1     X0611           ; jump if user flag 1 set (serial activity)
        mov     r0,#iostat
        mov     a,@r0
        anl     a,#20h          ; sound to be turned off ?
        jz      X060f           ; no
        mov     r0,#ipcstat
        mov     a,@r0
        anl     a,#0fdh         ; beep status to off
        mov     @r0,a
X060f:  jmp     X0200           ; get IPC command

X0611:  mov     r0,#iostat
        mov     a,@r0
        anl     a,#08           ; stop bit received ?
        jz      X064a           ; yes (serial byte in progress)
        mov     r0,#bdrate      ; inactiviy timeout for baud rate
        mov     r1,#bdcount     ; timeout counter
        inc     @r1             ; increment for this call
        call    X0700           ; is @r0 > @r1 ?
        jnc     X064a           ; yes, not timed out
        mov     r0,#ipcstat
        mov     a,@r0
        anl     a,#80h          ; else already timed out ?
        jz      X063e           ; no, first timeout call
        clr     f0              ; clear flag 0, ignore interrupts
        mov     a,@r0
        anl     a,#7fh          ; clear timed out flag
        mov     @r0,a
        stop    tcnt            ; stop timer/counter
        dis     i               ; disable external interrupts
        dis     tcnti           ; disable timer interrupts
        cpl     f1              ; clear user flag 1 (no serial activity)
        sel     rb1             ; switch to register bank 1
        mov     r0,#4bh         ; current serial queue pointer
        mov     a,@r0
        mov     r1,a            ; point to actual queue
        mov     a,r2            ; byte received
        mov     @r1,a           ; put into current queue
        sel     rb0             ; restore to register bank 0
        call    X004a           ; do housekeeping (if required, switch ports)
        cpl     f0              ; flag 0 set, service interrupts
        jmp     X0200           ; get IPC command

X063e:  orl     p2,#30h         ; busy both SER1 CTS & SER2 DTR
        mov     r0,#ipcstat
        mov     a,@r0
        orl     a,#80h          ; signal timeout occured
        mov     @r0,a
        mov     @r1,#0          ; clear timeout counter (or active queue size!)
        jmp     X0200           ; get IPC command

X064a:  mov     r1,#23h         ; currently active queue size
        mov     a,@r1
        add     a,#0e9h         ; >=23 queue full ?
        jc      X063e           ; yes
        mov     r0,#iostat
        mov     a,@r0
        anl     a,#2            ; current queue full ?
        jnz     X063e           ; yes
        jmp     X0200           ; else get IPC command

        .org    700h            ; 65A-6FF = 166 bytes spare

; compare bytes at @r0 and @r1 (carry set if @r0 <= @r1)

X0700:  mov     a,@r0
        jz      X0707
        cpl     a
        inc     a               ; form two's complement
        add     a,@r1
        ret

X0707:  mov     a,@r1
        clr     c               ; clear carry
        cpl     c               ; set carry, @r0 <= @r1
        ret

; compare words at @r0 and @r1 (carry set if @r0 <= @r1)

X070b:  call    X0700           ; do msb's @r0 = @r1 ?
        jnz     X0715           ; no
        dec     r0
        dec     r1
        call    X0700           ; do lsb's at @r0 = @r1 ?
        inc     r0              ; (flags as result)
        inc     r1
X0715:  ret                     ; flags unchanged on return

; add A to @r1 and any carry to @r1+1  (preserve r1)

X0716:  add     a,@r1
        mov     @r1,a
        inc     r1
        mov     a,#0
        addc    a,@r1
        mov     @r1,a
        dec     r1
        ret

; perform r2 * r4:r3 result to r6:r5

X071f:  mov     r5,#0           ; result
        mov     r6,#0
        clr     c               ; clear carry
        mov     r0,#8           ; iterations
X0726:  mov     a,r2
        rrc     a               ; has bit any weight ?
        mov     r2,a
        jnc     X0731           ; none
        mov     a,r3
        add     a,r5            ; else add to result
        mov     r5,a
        mov     a,r4
        addc    a,r6
        mov     r6,a
X0731:  mov     a,r3
        rlc     a               ; update weight for bit
        mov     r3,a
        mov     a,r4
        rlc     a
        mov     r4,a
        djnz    r0,X0726        ; until done
        ret

; randomise IPC random number (for beep or IPC command E)

X073a:  mov     r1,#26h         ; random number msb
        mov     a,@r1
        mov     r2,a
        mov     r3,#8dh         ; constant (r4 value as when called)
        call    X071f           ; perform r2 * r4:r3 result to r6:r5
        mov     a,r6
        mov     r1,#26h
        mov     @r1,a           ; update random number msb
        dec     r1              ; point to random number lsb
        mov     a,r5
        call    X0716           ; add A to word at @r1
        mov     a,#8bh
        call    X0716           ; add A to word at @r1
        ret

; fetch 4 bit data from ZX8302 to A lower nibble

X074f:  mov     r6,#0           ; clear data register
        mov     r7,#4           ; bits to fetch
X0753:  in      a,p2
        anl     a,#80h          ; COMDATA low ?
        jnz     X0753           ; no, wait until low
        movx    @r0,a           ; pulse COMCTL low (WR line)
        in      a,p2            ; read COMDATA
        movx    @r0,a           ; pulse COMCTL low (WR line)
        rlc     a               ; save data bit in carry
        mov     a,r6            ; accumualted data
        rlc     a               ; add new data bit
        mov     r6,a            ; preserve
        djnz    r7,X0753        ; until all bits done
        retr                    ; restore status and return

; send upper nibble of A to ZX8302

X0762:  mov     r7,#4           ; bits to send
        mov     r6,a            ; data to send
X0765:  in      a,p2
        anl     a,#80h          ; COMDATA low ?
        jnz     X0765           ; no, wait until low
        mov     a,r6            ; get current data bit
        movx    @r0,a           ; pulse COMCTL low (WR line)
        jb7     X0772           ; jump if data bit was high
        anl     p2,#7fh         ; else set COMDATA low
        jmp     X0774

X0772:  orl     p2,#80h         ; set COMDATA high
X0774:  movx    @r0,a           ; pulse COMCTL low (WR line)
        rl      a               ; next bit
        mov     r6,a            ; preserve
        orl     p2,#80h         ; restore COMDATA high
        djnz    r7,X0765        ; until all bits done
        retr                    ; restore status and return

; fetch 8 bit data from ZX8302 to A

X077c:  call    X074f           ; fetch nibble
        swap    a
        mov     r5,a            ; preserve
        call    X074f           ; fetch nibble
        orl     a,r5            ; form 8 bit data
        retr                    ; restore status and return

; send 8 bit data in A to ZX8302

X0784:  call    X0762           ; send nibble
        call    X0762           ; send nibble
        retr                    ; restore status and return

; send lower nibble of A to ZX8302

X0789:  swap    a               ; into suitable position
        call    X0762           ; send nibble
        retr                    ; restore status and return

        .org    800h            ; 78D-7FFh = 115 bytes spare
        .exit

